from libtuer import ThreadFunction, logger, fire_and_forget
from actor import Actor
import os, random

# logger.{debug,info,warning,error,critical}

def play_sound (what):
	try:
		soundfiles = os.listdir(SOUNDS_DIRECTORY+what)
	except FileNotFoundError:
		logger.error("StateMachine: Unable to list sound files in %s" % (SOUNDS_DIRECTORY+what))
		return
	soundfile = SOUNDS_DIRECTORY + what + '/' + random.choice(soundfiles)
	fire_and_forget ([SOUNDS_PLAYER,soundfile], logger.error, "StateMachine: ")


# StateOpening constants
OPEN_REPEAT_TIMEOUT = 8
OPEN_REPEAT_NUMBER = 3

# StateAboutToOpen constants
ABOUTOPEN_NERVLIST = [(5, lambda : play_sound("flipswitch")), (10, lambda:play_sound("flipswitch")), (10, lambda:Logger.warning("Space open but switch not flipped for 10 seconds")),\
	(20, lambda:play_sound("flipswitch")), (30, lambda:play_sound("flipswitch")), (30, lambda:Logger.error("Space open but switch not flipped for 30 seconds")),\
	(40, lambda:play_sound("flipswitch")), (50, lambda:play_sound("flipswitch")), (56, lambda:play_sound("flipswitch")), (60, lambda:Logger.critical("Space open but switch not flipped for 60 seconds"))]

# play_sound constants
SOUNDS_DIRECTORY = "/opt/tuer/sounds/"
SOUNDS_PLAYER = "/usr/bin/mplayer"


class StateMachine():
	# commands you can send
	CMD_PINS = 0
	CMD_BUZZ = 1
	CMD_OPEN = 2
	CMD_WAKEUP = 3
	CMD_LAST = 4
	
	class State():
		def __init__(self, state_machine):
			self.state_machine = state_machine
			self.time_entered = time.time()
			self.theDict = None
		def handle_pins_event(self):
			pass # one needn't implement this
		def handle_buzz_event(self,arg): # this shouldn't be overwritten
			self.actor.act(Actor.CMD_BUZZ)
			arg("200 okay: buzz executed")
		def handle_open_event(self,arg):
			if arg is not None:
				arg("412 Precondition Failed: The current state (%s) cannot handle the OPEN event" % self.__class__.__name__)
		def handle_wakeup_event(self):
			pass # one needn't implement this
		def pins(self):
			return self.state_machine.pins
		def actor(self):
			return self.state_machine.actor
		def handle_event(self,ev,arg):
			if arg is CMD_PINS:
				self.handle_pins_event()
			elif arg is CMD_BUZZ:
				self.handle_buzz_event(arg)
			elif arg is CMD_OPEN:
				self.handle_open_event(arg)
			elif arg is CMD_WAKEUP:
				self.handle_wakeup_event()
			else:
				raise Exception("Unknown command number: %d" % ev)
	
	class StateStart(State):
		def __init__(self, sm):
			State.__init__(self,sm)
		def handle_pins_event(self):
			thepins = self.pins()
			for pin in thepins:
				if pin is None:
					return None
			if thepins.door_locked:
				return StateZu
			else:
				return StateAuf

	class StateZu(State):
		def __init__(self,sm):
			State.__init__(self,sm)
		def handle_pins_event(self):
			pins = self.pins()
			if not pins.door_locked:
				return StateAboutToOpen(self.state_machine)
		def handle_open_event(self,callback):
			return StateOpening(callback,self.state_machine)
	
	class StateOpening(State):
		def __init__(self,callback,sm):
			State.__init__(self,sm)
			self.callbacks=[callback]
			self.tries = 0
			self.actor().act(Actor.CMD_OPEN)
		def notify(self, did_it_work):
			s = "200 okay: door open" if did_it_work else ("500 internal server error: Couldn't open door with %d tries à %f seconds" % (OPEN_REPEAT_NUMBER,OPEN_REPEAT_TIMEOUT))
			for cb in self.callbacks:
				if cb is not None:
					cb(s)
		def handle_pins_event(self):
			pins = self.pins()
			if not pins.door_locked:
				self.notify(True)
				return StateAboutToOpen(self.state_machine)
		def handle_open_event(self,callback):
			self.callbacks.append(callback)
		def handle_wakeup_event(self):
			over = time.time() - self.time_entered
			nexttry = (self.tries+1) * OPEN_REPEAT_TIMEOUT
			if over > nexttry:
				if self.tries < OPEN_REPEAT_NUMBER:
					self.actor().act(Actor.CMD_OPEN)
					self.tries += 1
				else:
					#TODO: LOG ERROR und EMAIL an Admins
					self.notify(False)
					return StateZu(self.state_machine)
	
	class StateAboutToOpen(State):
		def __init__(self, sm):
			State.__init__(sm)
		def handle_pins_event(self):
			pins = self.pins()
			if pins.door_locked:
				return StateZu(self.state_machine)
			elif pins.space_active:
				return StateAuf(self.state_machine)
			else:
				over = time.time() - self.time_entered
				# TODO: Nerv
				logger.debug("AboutToOpen since %f seconds. TODO: nerv the user" % over)
		# TODO Conny
	
	class StateAuf(State):
		#TODO Conny
		pass
	
	class StateClosing(State):
		#TODO Ralf/Conny
		pass
	
	class StateAboutToLeave(State):
		#TODO Ralf
		pass
	
	class StateLeaving(State):
		#TODO Ralf
		pass
	
	def __init__(self, actor):
		self.actor = actor
		self.callback = ThreadFunction(self._callback)
		self.current_state = None
		self.pins = None
	
	def stop (self):
		self.callback.stop()
	
	def _callback(self, cmd, arg=None):
		# update pins
		if cmd == StateMachine.CMD_PINS:
			self.pins = arg
		# handle stuff
		newstate = self.current_state.handle_event(cmd,arg) # returns None or an instance of the new state
		while newstate is not None:
			logger.info("StateMachine: new state = %s" % newstate.__class__.__name__)
			self.current_state = newstate
			newstate = self.current_state.handle_event(StateMachine.CMD_PINS, self.pins)
